﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.VehicleRouting.Interfaces;
using HeuristicLab.Problems.VehicleRouting.ProblemInstances;
using HeuristicLab.Problems.VehicleRouting.Variants;
using HeuristicLab.Random;

namespace HeuristicLab.Problems.VehicleRouting.Encodings.Potvin {
  [Item("ClusterCreator", "Creates a VRP solution by clustering customers first with a KMeans-algorithm and building tours afterwards alternatevly in a random or a greedy fashion.")]
  [StorableClass]
  public sealed class ClusterCreator : PotvinCreator, IStochasticOperator {

    public ILookupParameter<IRandom> RandomParameter {
      get { return (LookupParameter<IRandom>)Parameters["Random"]; }
    }

    public IValueParameter<IntValue> MinK {
      get { return (IValueParameter<IntValue>)Parameters["MinK"]; }
    }
    public IValueParameter<IntValue> MaxK {
      get { return (IValueParameter<IntValue>)Parameters["MaxK"]; }
    }

    public IValueParameter<DoubleValue> ClusterChangeThreshold {
      get { return (IValueParameter<DoubleValue>)Parameters["ClusterChangeThreshold"]; }
    }

    public IValueParameter<DoubleArray> TourCreationProbabilities {
      get { return (IValueParameter<DoubleArray>)Parameters["TourCreationProbabilities"]; }
    }

    [StorableConstructor]
    public ClusterCreator(bool deserializing) : base(deserializing) { }

    public ClusterCreator() : base() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The pseudo random number generator."));
      Parameters.Add(new ValueParameter<IntValue>("MinK", "Minimum number of clusters which should be created", new IntValue(1)));
      Parameters.Add(new OptionalValueParameter<IntValue>("MaxK", "Maximum number of clusters which should be created"));
      Parameters.Add(new ValueParameter<DoubleValue>("ClusterChangeThreshold", "A certain ratio below which a cluster should be considered as computed satisfactory", new DoubleValue(0.0)));     
      Parameters.Add(new ValueParameter<DoubleArray>("TourCreationProbabilities", "This weightage array denotes the probability which tour creation method is employed: [shuffled; greedy]", new DoubleArray(new double[] {1, 1})));
    }

    private ClusterCreator(ClusterCreator original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ClusterCreator(this, cloner);
    }

    public static PotvinEncoding CreateSolution(IVRPProblemInstance instance, IRandom random, int minK, int maxK, double clusterChangeThreshold, int creationOption) {
      PotvinEncoding result = new PotvinEncoding(instance);

      IPickupAndDeliveryProblemInstance pdp = instance as IPickupAndDeliveryProblemInstance;

      // add all customers
      List<int> customers = new List<int>();
      for (int i = 1; i <= instance.Cities.Value; i++) {
        if(pdp == null || pdp.GetDemand(i) >= 0)
          customers.Add(i);
      }

      // cluster by spatial distribution

      // (1) wrap stops in ClusterObject
      List<ClusterObject> clusterObjects = new List<ClusterObject>();
      foreach (int customer in customers) {
        clusterObjects.Add(new ClusterObject(customer, instance.GetCoordinates(customer)));
      }

      // (2) cluster
      int k = random.Next(minK, maxK);
      List<Cluster> clusters = KMeans(instance, random, clusterObjects, k, clusterChangeThreshold);

      // (3) build tours
      // (a) shuffle
      // (b) greedy
      foreach (Cluster c in clusters) {
        Tour newTour = new Tour();
        result.Tours.Add(newTour);

        if (creationOption == 0) {
          // (a) shuffle
          c.Objects.Shuffle(random);
          foreach (ClusterObject o in c.Objects) {
            newTour.Stops.Add(o.Id);
          }
        } else {
          // (b) greedy
          foreach (ClusterObject o in c.Objects) {
            newTour.Stops.Add(o.Id);
          }
          GreedyTourCreation(instance, result, newTour, false);
        }
      }
      return result;
    }

    public override IOperation InstrumentedApply() {
      IRandom random = RandomParameter.ActualValue;

      int minK = (MinK.Value.Value > 0) ? MinK.Value.Value : 1;
      int maxK = (MaxK.Value != null) ? MaxK.Value.Value : ProblemInstance.Vehicles.Value;
      double clusterChangeThreshold = (ClusterChangeThreshold.Value.Value >= 0.0 &&
                                       ClusterChangeThreshold.Value.Value <= 1.0)
        ? ClusterChangeThreshold.Value.Value
        : 0.0;

      // normalize probabilities
      double max = TourCreationProbabilities.Value.Max();
      double[] probabilites = new double[2];
      for (int i = 0; i < TourCreationProbabilities.Value.Length; i++) {
        probabilites[i] = TourCreationProbabilities.Value[i] / max;
      }

      List<int> creationOptions = new List<int>() {0, 1};
      int creationOption = creationOptions.SampleProportional(random, 1, probabilites, false, false).First();

      VRPToursParameter.ActualValue = CreateSolution(ProblemInstance, random, minK, maxK, clusterChangeThreshold, creationOption);
      return base.InstrumentedApply();
    }

    private static List<Cluster> KMeans(IVRPProblemInstance instance, IRandom random, List<ClusterObject> objects, int k, double changeTreshold) {

      List<Cluster> clusters = new List<Cluster>();
      HashSet<int> initMeans = new HashSet<int>();
      int nextMean = -1;

      // (1) initialize each cluster with a random first object (i.e. mean)
      for (int i = 0; i < k && i < objects.Count; i++) {
        Cluster cluster = new Cluster(instance, i);

        do {
          nextMean = random.Next(0, objects.Count);
        } while (initMeans.Contains(nextMean));
        initMeans.Add(nextMean);
        cluster.SetMean(objects[nextMean]);
        clusters.Add(cluster);
      }

      // (2) repeat clustering until change rate is below threshold
      int changes = 0;
      double changeRate = 1.0;

      do {
        changes = KMeansRun(clusters, objects);
        changeRate = changes / objects.Count;

      } while (changeRate > changeTreshold);

      // remove empty clusters
      clusters.RemoveAll(c => c.Objects.Count.Equals(0));

      return clusters;
    }

    private static int KMeansRun(List<Cluster> clusters, List<ClusterObject> objects) {
      int changes = 0;

      foreach (Cluster c in clusters) {
        c.Objects.Clear();
      }

      foreach (ClusterObject o in objects) {
        int bestClusterIdx = -1;
        double minImpact = double.MaxValue;
        for(int i = 0; i < clusters.Count; i++) {
          double impact = clusters[i].CalculateVarianceIncrease(o);
          if (impact < minImpact) {
            minImpact = impact;
            bestClusterIdx = i;
          }
        }
        clusters[bestClusterIdx].AddObject(o);
      }

      foreach (Cluster c in clusters) {
        c.CalculateMean();
      }

      return changes;
    }

    class Cluster {
      public int Id;
      public List<ClusterObject> Objects { get; private set; }
      public double Variance { get; private set; }
      public ClusterObject Mean { get; set; }

      private readonly IVRPProblemInstance instance;

      public Cluster(IVRPProblemInstance instance, int id) {
        this.instance = instance;
        Id = id;
        Objects = new List<ClusterObject>();
      }

      public bool IsEqual(Cluster cluster) {
        bool equal = (cluster != null) && (cluster.Objects.Count == Objects.Count);
        int index = 0;

        while (equal && index < Objects.Count) {
          equal = cluster.Objects.Contains(Objects[index]);
          index++;
        }
        return equal;
      }

      public bool AddObject(ClusterObject o) {
        Objects.Add(o);

        bool clusterChanged = o.ClusterId != Id;
        o.ClusterId = Id;
        return clusterChanged;
      }

      public void SetMean(ClusterObject o) {
        Mean = o;
        Mean.ClusterId = Id;
      }

      public ClusterObject CalculateMean() {
        int dimensions = instance.Coordinates.Columns;
        ClusterObject mean = new ClusterObject(-1, Id, dimensions);
        foreach (ClusterObject o in Objects) {
          double[] coordinates = instance.GetCoordinates(o.Id);
          for (int i = 0; i < coordinates.Length; i++) {
            mean.Coordinates[i] += (coordinates[i] / Objects.Count);
          }
        }
        Mean = mean;
        return Mean;
      }

      public double CalculateVariance() {
        Variance = 0.0;

        if (Mean == null)
          CalculateMean();

        foreach (ClusterObject o in Objects) {
          double[] coordinates = instance.GetCoordinates(o.Id);
          Variance += Math.Pow(Mean.GetDistance(o.Coordinates), 2);
        }
        return Variance /= Objects.Count;
      }

      public double CalculateVarianceIncrease(ClusterObject o) {
        if(Mean == null)
          CalculateMean();

        return Math.Pow(Mean.GetDistance(o.Coordinates), 2) / ((Objects.Count > 0) ? Objects.Count : 1);
      }
    }

    class ClusterObject {
      
      public int Id { get; private set; }

      public int ClusterId { get; set; }

      public double[] Coordinates { get; set; }

      public ClusterObject() { }

      public ClusterObject(int id, int clusterId = -1) {
        Id = id;
        ClusterId = clusterId;
      }

      public ClusterObject(int id, double[] coordinates) {
        Id = id;
        Coordinates = coordinates;
      }

      public ClusterObject(int id, int clusterId, int dimensions) {
        Id = id;
        ClusterId = clusterId;
        Coordinates = new double[dimensions];
      }

      public double GetDistance(double[] otherCoordinates) {
        double distance = 0.0;

        for (int i = 0; i < otherCoordinates.Length; i++) {
          distance += Math.Pow(Coordinates[i] - otherCoordinates[i], 2);
        }
        return Math.Sqrt(distance);
      }
    }

    private static double CalculateAngleToDepot(IVRPProblemInstance instance, int city) {
      double dx = instance.GetCoordinates(0)[0];
      double dy = instance.GetCoordinates(0)[1];

      double cx = instance.GetCoordinates(city)[0];
      double cy = instance.GetCoordinates(city)[1];

      double alpha = Math.Atan((cx - dx) / (dy - cy)) * (180.0 / Math.PI);
      if (cx > dx && cy > dy)
        alpha = (90.0 + alpha) + 90.0;
      else if (cx < dx && cy > dy)
        alpha = alpha + 180.0;
      else if (cx < dx && cy < dy)
        alpha = (90.0 + alpha) + 270.0;

      return alpha;
    }

    private static void GreedyTourCreation(IVRPProblemInstance instance, PotvinEncoding individual, Tour tour, bool includeCosts) {
      Tour unrouted = tour.Clone() as Tour;
      tour.Stops.Clear();
      VRPEvaluation evaluation = instance.EvaluateTour(tour, individual);
      double currentQuality = (includeCosts) ? evaluation.Quality : evaluation.Distance;

      while (unrouted.Stops.Count > 0) {
        int insertionPosition = tour.Stops.Count;
        int nextCityIdxToAdd = 0;

        tour.Stops.Add(unrouted.Stops[nextCityIdxToAdd]);
        evaluation = instance.EvaluateTour(tour, individual);
        double minQualityIncrease = ((includeCosts) ? evaluation.Quality : evaluation.Distance) - currentQuality;
        tour.Stops.RemoveAt(insertionPosition);

        for (int j = 1; j < unrouted.Stops.Count; j++) {
          tour.Stops.Insert(insertionPosition, unrouted.Stops[j]);
          evaluation = instance.EvaluateTour(tour, individual);
          double qualityIncrease = ((includeCosts) ? evaluation.Quality : evaluation.Distance) - currentQuality;
          tour.Stops.RemoveAt(insertionPosition);

          if (qualityIncrease < minQualityIncrease) {
            nextCityIdxToAdd = j;
            minQualityIncrease = qualityIncrease;
          }
        }

        currentQuality += minQualityIncrease;
        tour.Stops.Insert(insertionPosition, unrouted.Stops[nextCityIdxToAdd]);
        unrouted.Stops.RemoveAt(nextCityIdxToAdd);
      }
    }

  }
}